#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import logging

from django.conf import settings
from django.core.cache import cache
from oauthlib.oauth1 import RequestValidator

from django_declarative_apis import models as oauth_models


logger = logging.getLogger(__name__)


class DjangoRequestValidator(RequestValidator):

    TIMESTAMP_THRESHOLD = 300

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.consumer = None
        self.validation_error_message = ""

    @property
    def enforce_ssl(self):
        return getattr(settings, "REQUIRE_HTTPS_FOR_OAUTH", True)

    def check_client_key(self, client_key):
        return bool(client_key)

    def check_nonce(self, nonce):
        return bool(
            nonce
        )  # we didn't enforce any minimum length on nonces previously...

    def validate_timestamp_and_nonce(
        self,
        client_key,
        timestamp,
        nonce,
        request,
        request_token=None,
        access_token=None,
    ):
        request_token = request_token or ""
        access_token = access_token or ""
        cache_key = "{0}:{1}:{2}:{3}".format(
            client_key, request_token, access_token, nonce
        )
        cache_created = cache.add(cache_key, True, self.TIMESTAMP_THRESHOLD * 2)
        if not cache_created:
            self.validation_error_message = "nonce_used"
        return cache_created

    def validate_client_key(self, client_key, request):
        self.consumer = oauth_models.get_consumer(client_key)
        if self.consumer is None:
            self.validation_error_message = "consumer_key_unknown"
            logger.error("invalid consumer")
            return False
        return True

    def get_client_secret(self, client_key, request):
        try:
            if self.consumer.rsa_public_key_pem:
                return None
            else:
                return self.consumer.secret
        except Exception as e:  # noqa
            logger.error(
                "This should never happen, since consumer is already validated"
            )
            return ""

    def get_rsa_key(self, client_key, request):
        try:
            return self.consumer.rsa_public_key_pem
        except Exception as e:  # noqa
            logger.error(
                "This should never happen, since consumer is already validated"
            )
            return ""

    @property
    def dummy_client(self):
        class DummyClient:
            def __init__(self, *args, **kwargs):
                self.secret = ""
                self.key = ""
                self.rsa_public_key_base64 = ""

        return DummyClient()



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

from collections import defaultdict
import inspect
import logging
import types

import pydantic
from django.conf import settings
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import ManyToOneRel
from django.db.models.fields.reverse_related import ForeignObjectRel

logger = logging.getLogger(__name__)

NEVER = 0
ALWAYS = 1
IF_TRUTHY = 2

DEFAULT_UNEXPANDED_VALUE = object()
EXPANDABLE_FIELD_KEY = "__expandable__"


class _ExpandableForeignKey:
    def __init__(self, display_key, model_class, inst_field_name):
        self.display_key = display_key
        self.model_class = model_class
        self.inst_field_name = inst_field_name


def expandable(model_class=None, display_key=None, inst_field_name=None):
    if model_class and not issubclass(model_class, (models.Model,)):
        raise ValueError("model_class must be an instance of a Django Model")
    if model_class and display_key:
        try:
            model_class._meta.get_field(display_key)
        except FieldDoesNotExist:
            raise ValueError(f"{display_key} is not a field on {model_class.__name__}")
    return _ExpandableForeignKey(display_key, model_class, inst_field_name)


def _get_unexpanded_field_value(inst, field_name, field_type):
    if not field_type.model_class:
        return DEFAULT_UNEXPANDED_VALUE

    display_key = field_type.display_key or field_type.model_class._meta.pk.name
    inst_field_name = field_type.inst_field_name or field_name
    is_multiple = isinstance(
        inst.__class__._meta.get_field(inst_field_name), ManyToOneRel
    )

    if is_multiple:
        # special case for keys that have multiple values (for instance, inverse fk relations)
        obj = getattr(inst, inst_field_name)
        return [{display_key: getattr(v, display_key)} for v in obj.all()]

    if display_key == field_type.model_class._meta.pk.name:
        # special case - we know this is a primary key, so we can get it without retrieving the object
        return {display_key: getattr(inst, inst_field_name + "_id")}
    else:
        # we're not returning the PK - have to actually retrieve the model
        obj = getattr(inst, inst_field_name)
        return {display_key: getattr(obj, display_key)}


def _is_relation(field):
    return field.is_relation or getattr(field, "is_fake_relation", False)


def _is_reverse_relation(field):
    return isinstance(field, ForeignObjectRel)


def _get_callable_field_value_with_cache(inst, field_name, model_cache, field_type):
    cache_relation = True

    try:
        field_meta = inst._meta.get_field(field_name)
        if not _is_relation(field_meta) or _is_reverse_relation(field_meta):
            # no `attname` in reverse relations
            cache_relation = False
    except (AttributeError, FieldDoesNotExist):
        # inst doesn't look like a django model
        cache_relation = False

    if cache_relation:
        # we're caching a foreign key field on a django model.  Cache it by (model, fk_pk) so that if
        # other objects reference this same instance, we'll get a cache hit
        fk_pk = getattr(inst, field_meta.attname)
        val_cls = field_meta.related_model
        cache_key = _make_model_cache_key(val_cls, fk_pk)
    else:
        # not a foreign key.  Cache it by (inst, field_name) - it won't be a cache hit on another instance, but
        # will be cached if this same inst is returned later in the response
        cache_key = (id(inst), field_name)

    if cache_key in model_cache:
        logger.debug("ev=model_cache, status=hit, key=%s", cache_key)
        result = model_cache[cache_key]
    else:
        logger.debug("ev=model_cache, status=miss, key=%s", cache_key)
        result = field_type(inst)

        if isinstance(result, models.Manager):
            # need to get an iterable to proceed
            result = result.all()
        model_cache[cache_key] = result
    return result


def _get_filtered_field_value(  # noqa: C901
    inst,
    field_name,
    field_type,
    filter_def,
    expand_this,
    expand_children,
    filter_cache,
    model_cache,
):
    # get the value from inst
    if field_type == NEVER:
        return None

    if isinstance(field_type, types.FunctionType):
        if is_caching_enabled():
            val = _get_callable_field_value_with_cache(
                inst, field_name, model_cache, field_type
            )
        else:
            val = field_type(inst)
    elif isinstance(field_type, _ExpandableForeignKey):
        if expand_this:
            inst_field_name = field_type.inst_field_name or field_name
            val = getattr(inst, inst_field_name)
        else:
            val = _get_unexpanded_field_value(inst, field_name, field_type)
    else:
        try:
            if isinstance(inst, (models.Model)):
                try:
                    field_meta = inst._meta.get_field(field_name)
                    if _is_relation(field_meta):
                        val_pk = getattr(inst, field_meta.attname)
                        val_cls = field_meta.related_model
                        val_expand_children = expand_children.get(field_name, {})
                        cache_key = _make_filter_cache_key(
                            val_expand_children, val_cls, val_pk
                        )
                        if cache_key in filter_cache:
                            logger.debug(
                                "ev=filter_cache, status=hit, key=%s", cache_key
                            )
                            return filter_cache[cache_key]
                except FieldDoesNotExist:
                    # this happens when you reference the special field "pk" in filters
                    pass

            val = getattr(inst, field_name)
        except (AttributeError, FieldDoesNotExist):
            return None

    if isinstance(val, models.Manager):
        # need to get an iterable to proceed
        val = val.all()

    # should this value be passed through the filters itself?
    # `dict` is intentionally excluded to allow endpoints to return an arbitrary response that bypasses filtering
    if val.__class__ in filter_def or isinstance(
        val, (list, tuple, models.Model, models.query.QuerySet, pydantic.BaseModel)
    ):
        val = _apply_filters_to_object(
            val, filter_def, expand_children, val.__class__, filter_cache, model_cache
        )

    if (
        (field_type == ALWAYS)
        or isinstance(field_type, types.FunctionType)
        or ((field_type == IF_TRUTHY) and val)
        or isinstance(field_type, _ExpandableForeignKey)
    ):
        return val
    else:
        return None


def _make_filter_cache_key(expand_children, klass, pk):
    return (str(expand_children), klass, str(pk))


def _make_model_cache_key(klass, pk):
    return (klass, str(pk))


# TODO: make this method less complex and remove the `noqa`
def _apply_filters_to_object(  # noqa: C901
    inst, filter_def, expand_children, klass, filter_cache, model_cache
):
    is_cacheable = False
    if (
        is_caching_enabled()
        and isinstance(inst, (models.Model,))
        and filter_cache is not None
    ):
        is_cacheable = True
        pk = getattr(inst, "pk")
        cache_key = _make_filter_cache_key(expand_children, klass, pk)
        if cache_key in filter_cache:
            logger.debug("ev=filter_cache, status=hit, key=%s", cache_key)
            return filter_cache[cache_key]
        else:
            logger.debug("ev=filter_cache, status=miss, key=%s", cache_key)
    if isinstance(inst, (list, tuple, models.query.QuerySet)):
        # if it's a tuple or list, iterate over the collection and call _apply_filters_to_object on each item
        return [
            _apply_filters_to_object(
                item,
                filter_def,
                expand_children,
                item.__class__,
                filter_cache,
                model_cache,
            )
            for item in inst
        ]
    elif isinstance(inst, (dict,)):
        return {
            k: _apply_filters_to_object(
                v,
                filter_def,
                expand_children,
                v.__class__,
                filter_cache,
                model_cache,
            )
            for k, v in inst.items()
        }

    fields_def = filter_def.get(klass)
    if fields_def is None:
        # no filter definition for this class.  Check the inheritance chain to see if we have something there
        result = None
        for base_class in inspect.getmro(klass):
            if base_class in filter_def:
                result = _apply_filters_to_object(
                    inst,
                    filter_def,
                    expand_children,
                    base_class,
                    filter_cache,
                    model_cache,
                )
                break

        if is_cacheable:
            filter_cache[cache_key] = result

        return result
    else:
        # first, recursively populate from any ancestor classes in the inheritance hierarchy
        result = defaultdict(list)
        for base in klass.__bases__:
            filtered_ancestor = _apply_filters_to_object(
                inst, filter_def, expand_children, base, filter_cache, model_cache
            )
            if filtered_ancestor:
                result.update(filtered_ancestor)

        # now populate the fields according to fields defined in filter_def
        expandables = []
        if isinstance(fields_def, types.FunctionType):
            result = fields_def(inst)
        elif fields_def is ALWAYS:
            return inst
        else:
            for field_name, field_type in fields_def.items():
                if isinstance(field_type, _ExpandableForeignKey):
                    expandables.append(field_name)

                value = _get_filtered_field_value(
                    inst,
                    field_name,
                    field_type,
                    filter_def,
                    expand_this=field_name in expand_children,
                    expand_children=expand_children.get(field_name, {}),
                    filter_cache=filter_cache,
                    model_cache=model_cache,
                )

                if value is not None and value != DEFAULT_UNEXPANDED_VALUE:
                    result[field_name] = value
                elif (field_type == NEVER) and (field_name in result):
                    del result[field_name]

        if expandables:
            result[EXPANDABLE_FIELD_KEY] += expandables

        if is_cacheable:
            filter_cache[cache_key] = result

        return result


def _compile_expansion(expand_fields):
    top = {}
    for field in expand_fields:
        d = top
        for key in field.strip().split("."):
            d.setdefault(key, {})
            d = d[key]
    return top


def is_caching_enabled():
    return getattr(settings, "DDA_FILTER_MODEL_CACHING_ENABLED", False)


def apply_filters_to_object(inst, filter_def, expand_header=""):
    if expand_header:
        expand_dict = _compile_expansion(expand_header.split(","))
    else:
        expand_dict = {}
    return _apply_filters_to_object(
        inst,
        filter_def,
        expand_children=expand_dict,
        klass=inst.__class__,
        filter_cache={},
        model_cache={},
    )



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import importlib
import random as python_random
import string

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models as django_models
from django.conf import settings

KEY_SIZE = 18
SECRET_SIZE = 32

random = python_random.SystemRandom()


def get_consumer(key):
    getter_str = getattr(settings, "DECLARATIVE_ENDPOINT_CONSUMER_GETTER", None)
    if getter_str is None:
        try:
            return OauthConsumer.objects.get(key=key)
        except OauthConsumer.DoesNotExist as e:  # noqa
            return None
    else:
        module, getter = getter_str.rsplit(".", 1)
        module = importlib.import_module(module)
        getter = getattr(module, getter)
        if not callable(getter):
            raise TypeError(f"Consumer getter ({getter_str}) must be callable")
        return getter(key)


def get_random_string(
    length=20, chars=string.ascii_lowercase + string.ascii_uppercase + string.digits
):
    return "".join(random.choice(chars) for _ in range(length))


class BaseConsumer(django_models.Model):
    """:code:`BaseConsumer` is shared in common by all consumers. It implements Django’s :code:`GenericForiegnKey`
    to find what the consumer is supposed to point to, which could be the service calling the API to perform
    authentication, or it could be a mobile app instance.
    You can also set the read and write privileges of the consumer using TYPE_READ_ONLY and TYPE_READ_WRITE.
    """

    class Meta:
        abstract = True

    name = django_models.CharField(max_length=100, null=True, blank=True)

    content_type = django_models.ForeignKey(
        ContentType, on_delete=django_models.CASCADE, null=True
    )
    object_id = django_models.CharField(max_length=100, null=True)
    associated_object = GenericForeignKey()

    TYPE_READ_ONLY = "RO"
    TYPE_READ_WRITE = "RW"
    TYPES = ((TYPE_READ_ONLY, "Read Only"), (TYPE_READ_WRITE, "Read/Write"))
    type = django_models.CharField(
        max_length=2, choices=TYPES, default=TYPE_READ_WRITE, db_index=True
    )


class OauthConsumerManager(django_models.Manager):
    def create(self, **kwargs):
        key = kwargs.pop("key", get_random_string(length=KEY_SIZE))
        secret = kwargs.pop("secret", get_random_string(length=SECRET_SIZE))
        return super().create(key=key, secret=secret, **kwargs)


class OauthConsumer(BaseConsumer):
    """:code:`OAuthConsumer` inherits from :code:`BaseConsumer` and it based on :code:`OAuth1.0a`.
    It adds the additional properties of key, secret, and rsa_public_key_pem.

    **Example**

    .. code-block:: python

        from django_declarative_apis import models

        consumer = models.OAuthConsumer.objects.create()

        # *consumer will have:*
        # consumer.content_type_id
        # consumer.id
        # consumer.key
        # consumer.name
        # consumer.object_id
        # consumer.secret
        # consumer.type
    """

    objects = OauthConsumerManager()

    key = django_models.CharField(max_length=KEY_SIZE, db_index=True)
    """ Consumer key as defined by OAuth 1.0a
    """
    secret = django_models.CharField(max_length=SECRET_SIZE)
    rsa_public_key_pem = django_models.TextField(blank=True, null=True)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import time

from django_declarative_apis.authentication import AuthenticationFailure


class OAuthError(AuthenticationFailure):
    detail = None
    auth_header = None


class OAuthTimestampError(OAuthError):
    def __init__(self):
        self.detail = (
            "There was a problem with your timestamp. Please check your current system time."
            "Server time is {0}.".format(time.time())
        )
        start_time = int(time.time()) - 300
        end_time = int(time.time())
        self.auth_header = (
            'OAuth realm="API",oauth_problem=timestamp_refused'
            f"&oauth_acceptable_timestamps={start_time}-{end_time}"
        )


class OAuthMissingParameterError(OAuthError):
    def __init__(self, detail):
        self.detail = "Parameters missing: {0}".format(detail.split(":")[1])
        self.auth_header = 'OAuth realm="API",oauth_problem=parameter_absent&oauth_parameters_absent={0}'.format(
            detail.split(":")[1]
        )


class OAuthInvalidSignatureError(OAuthError):
    def __init__(self, detail):
        self.detail = "{0}.".format(detail)
        self.auth_header = 'OAuth realm="API",oauth_problem=signature_invalid'


class OAuthInvalidVersionError(OAuthError):
    def __init__(self):
        self.detail = "OAuth version (oauth_version parameter) should be 1.0"
        self.auth_header = 'OAuth realm="API",oauth_problem=version_rejected&oauth_acceptable_versions=1.0-1.0'


class OAuthParameterRejectedError(OAuthError):
    def __init__(self, detail):
        self.detail = "{0}.".format(detail)
        self.auth_header = 'OAuth realm="API",oauth_problem=parameters_rejected'


"""
Map internal OAuth errors to those specified in the OAuth Problem Reporting proposal
http://wiki.oauth.net/w/page/12238543/ProblemReporting
"""

error_to_human_readable_message = {
    "nonce_used": "Nonce was already used",
    "signature_method_rejected": 'Signature method must be "HMAC-SHA1" or "RSA-SHA1',
    "consumer_key_unknown": "Invalid consumer key or secret",
}


def build_error(error_message):
    if "Malformed authorization header" in error_message:
        return OAuthParameterRejectedError(error_message)
    if ("Timestamp given is invalid" in error_message) or (
        "Invalid timestamp" in error_message
    ):
        return OAuthTimestampError()
    if "parameter_absent" in error_message:
        return OAuthMissingParameterError(error_message)
    if "Invalid signature" in error_message:
        return OAuthInvalidSignatureError(error_message)
    if "Invalid OAuth version" in error_message:
        return OAuthInvalidVersionError()

    detail = error_to_human_readable_message.get(error_message, error_message)
    auth_header = 'OAuth realm="API",oauth_problem={0}'.format(error_message)

    return OAuthError(detail=detail, auth_header=auth_header)



#
# Copyright (c) 2019, salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#

import abc
import typing


class AuthenticatorHint(typing.NamedTuple):
    """Tuple to provide hints for authentication implementations.

    header_hint: a string used to match the :code:`Authentication:` header.
    """

    header: str


class Authenticator(metaclass=abc.ABCMeta):
    """The base class for constructing an authenticator.

    The Authenticator class has two methods: :code:`is_authenticated` and
    :code:`challenge`. Both of these need to be overridden by the authenticator
    implementation that inherits from :code:`Authenticator` class. Otherwise,
    it will throw a :code:`NotImplementedError`.

    **Example**

    .. code-block::

        from django_declarative_apis.authentication import Authenticator

        class SampleAuthenticator(Authenticator):
            def is_authenticated(request):
                # authentication code

            def challenge(self, error):
                # challenge code
    """

    @abc.abstractmethod
    def is_authenticated(self, request):
        """Takes in the request as an argument and identifies whether the
        requester is valid.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def challenge(self, error):
        """Results in the challenge response sent to the user

        This should result in a django.http.HttpResponse that should include
        information through the WWW-Authenticate header around expectations.
        """
        raise NotImplementedError


class AuthenticationResult:
    """A class definition that takes in and stores the authentication header
    and details of the result.
    """

    def __init__(self, detail=None, auth_header=None):
        self.detail = detail
        self.auth_header = auth_header


class NoAuthentication(AuthenticationResult, Authenticator):
    """
    Authentication handler that always returns True, so no authentication is
    needed, nor initiated.

    .. note::
        **Important:** In this implementation the :code:`challenge` method is
        missing and must be implemented by the user. Otherwise, it will raise
        :code:`NotImplementedError`.
    """

    def is_authenticated(self, request):
        return True


class AuthenticationSuccess(AuthenticationResult):
    """An instance of :code:`AuthenticationResult` that returns :code:`True`.

    It can be used as a return response in an authenticator implementation.
    """

    def __bool__(self):
        return True


class AuthenticationFailure(AuthenticationResult):
    """An instance of :code:`AuthenticationResult` that returns :code:`False`.

    It can be used as a return response in an authenticator implementation.
    """

    def __bool__(self):
        return False


def validate_authentication_config(config):
    """Validate the computed configuration of authentication handlers

    The schema for the config is:

    {
        <AuthenticatorHint>: [<Authenticator>, <Authenticator>...],
        <AuthenticatorHint>: [<Authenticator>, <Authenticator>...],
    }

    AuthenticatorHints allow us to match Authorization headers for quick
    handler lookup. For example, if we want to use OAuth 1.0a, we could use an
    AuthenticatorHint.header value of 'OAuth ' as a key and a value of
    [django_declarative_apis.authentication.oauthilib.oauth1.TwoLeggedOauth1()].
    This will ensure that any time an `Authorization: OAuth ...` header is
    seen, the appropriate authenticator is used.

    If there are more complexities to the authenticator (i.e. OAuth 1.0a allows
    for the transport of credentials through header, request body or query
    parameters, catch-alls are allowed by using a key of `None`:

    {
        None: [<Authenticator>],
    }

    The catch-all authenticators are always executed after matched
    authenticators.

    Note: This may need to get smarter in the future but was kept simple
          intentionally as it's executed on every request.
    """
    assert isinstance(config, typing.Mapping)
    for hint, authenticators in config.items():
        if not isinstance(hint, (AuthenticatorHint, type(None))):
            raise TypeError(
                "Authenticator hint must be an instance of authentication.AuthenticatorHint or None"
            )

        assert isinstance(authenticators, (list, tuple))
        for authenticator in authenticators:
            if not isinstance(authenticator, Authenticator):
                raise TypeError(
                    "Authenticator must be an instance of authentication.Authenticator"
                )
